from decimal import Decimal
from typing import Union
from py_portfolio_index.models import Money


def print_per(input: Union[Money, Decimal]):
    if isinstance(input, Money):
        return f"{round(input.value*100,4)}%"
    return f"{round(input*100,4)}%"


def print_money(input: Union[Money, Decimal]):
    if isinstance(input, Money):
        return f"{round(input,2)}"
    return f"${round(input,2)}"



from py_portfolio_index.models import RealPortfolio, RealPortfolioElement, Money
from .base_portfolio import BaseProvider
from decimal import Decimal
from typing import Optional
from datetime import date, datetime, timezone, timedelta
from functools import lru_cache

from os import environ


class AlpacaProvider(BaseProvider):
    def __init__(
        self,
        key_id: str | None = None,
        secret_key: str | None = None,
        paper: bool = False,
    ):
        from alpaca.trading.client import TradingClient
        from alpaca.data.historical import StockHistoricalDataClient

        if not key_id:
            key_id = environ.get("ALPACA_API_KEY", None)
        if not secret_key:
            secret_key = environ.get("ALPACA_API_SECRET", None)
        if not (key_id and secret_key):
            raise ValueError(
                "Must provide key_id and secret_key or set environment variables ALPACA_API_KEY and ALPACA_API_SECRET "
            )
        self.trading_client = TradingClient(
            api_key=key_id, secret_key=secret_key, paper=False
        )
        self.historical_client = StockHistoricalDataClient(
            api_key=key_id, secret_key=secret_key
        )
        # tradeapi.REST(
        #     key_id=key_id, secret_key=secret_key, base_url=URL(TARGET_URL)
        # )
        BaseProvider.__init__(self)

    @lru_cache(maxsize=None)
    def _get_instrument_price(
        self, ticker: str, at_day: Optional[date] = None
    ) -> Optional[Decimal]:
        from alpaca.data.timeframe import TimeFrame, TimeFrameUnit
        from alpaca.data.requests import StockBarsRequest, StockLatestQuoteRequest

        if at_day:
            start = datetime(at_day.year, at_day.month, at_day.day, tzinfo=timezone.utc)
            end = start + timedelta(days=7)
            # end = datetime(at_day.year, at_day.month, at_day.day+7, hour=23, tzinfo=timezone.utc)
            raw = self.historical_client.get_stock_bars(
                StockBarsRequest(
                    symbol_or_symbols=ticker,
                    start=start,
                    end=end,
                    timeframe=TimeFrame(amount=1, unit=TimeFrameUnit.Day),
                    limit=100,
                    adjustment=None,
                    feed=None,
                )
                # [ticker],
                # timeframe=TimeFrame(amount=1, unit=TimeFrameUnit.Day),
                # start=start.isoformat(),
                # end=end.isoformat(),
            )
            # take the first day after target day
            return Decimal(raw[ticker][0].high)
        else:
            raw = self.historical_client.get_stock_latest_quote(
                StockLatestQuoteRequest(symbol_or_symbols=ticker, feed=None)
            )
            # if we don't have a value for the current day
            # expand out
            # this is requuired on weekends
            if not raw[ticker].ask_price:
                default = datetime.now(tz=timezone.utc) - timedelta(hours=1)
                start = default - timedelta(days=7)
                end = default
                raw = self.historical_client.get_stock_bars(
                    StockBarsRequest(
                        symbol_or_symbols=ticker,
                        start=start,
                        end=end,
                        timeframe=TimeFrame(amount=1, unit=TimeFrameUnit.Day),
                        feed=None,
                        adjustment=None,
                        limit=1000,
                    )
                    # [ticker],
                    # timeframe=TimeFrame(amount=1, unit=TimeFrameUnit.Day),
                    # start=start.isoformat(),
                    # end=end.isoformat(),
                )
                # take the first day after target day
                return Decimal(raw[ticker][0].high)
            return Decimal(raw[ticker].ask_price)

    def buy_instrument(self, ticker: str, qty: Decimal):
        from alpaca.trading.requests import MarketOrderRequest
        from alpaca.trading.enums import OrderSide, TimeInForce

        market_order_data = MarketOrderRequest(
            symbol=ticker,
            qty=float(qty),
            side=OrderSide.BUY,
            time_in_force=TimeInForce.DAY,
        )
        self.trading_client.submit_order(order_data=market_order_data)

        return True

    def get_unsettled_instruments(self):
        from alpaca.trading.requests import GetOrdersRequest, QueryOrderStatus

        open_orders = self.trading_client.get_orders(
            filter=GetOrdersRequest(status=QueryOrderStatus.OPEN)
        )
        return set([o.symbol for o in open_orders])

    def get_holdings(self):
        from decimal import Decimal

        my_stocks = self.trading_client.get_all_positions()

        if not my_stocks:
            return RealPortfolio(holdings=[])
        total_value = sum([Decimal(item.market_value) for item in my_stocks])
        out = [
            RealPortfolioElement(
                ticker=row.symbol,
                units=row.qty,
                value=Money(value=Decimal(row.market_value)),
                weight=Decimal(row.market_value) / total_value,
            )
            for row in my_stocks
        ]
        return RealPortfolio(holdings=out)



from py_portfolio_index.bin import INDEXES, STOCK_LISTS
from py_portfolio_index.constants import Logger
from py_portfolio_index.enums import PurchaseStrategy
from py_portfolio_index.operators import compare_portfolios
from py_portfolio_index.portfolio_providers.alpaca import AlpacaProviderLegacy
from py_portfolio_index.portfolio_providers.robinhood import RobinhoodProvider
from py_portfolio_index.portfolio_providers.alpaca_v2 import AlpacaProvider
from py_portfolio_index.config import get_providers

AVAILABLE_PROVIDERS = get_providers()

__version__ = "0.0.5"

__all__ = [
    "INDEXES",
    "STOCK_LISTS",
    "Logger",
    "compare_portfolios",
    "AlpacaProvider",
    "AlpacaProviderLegacy",
    "RobinhoodProvider",
    "PurchaseStrategy",
]



from datetime import datetime, date
from typing import List, Optional, Union, TYPE_CHECKING
from pydantic import BaseModel, Field, validator
from pandas import DataFrame
from py_portfolio_index.enums import Currency
from py_portfolio_index.constants import Logger
from py_portfolio_index.exceptions import PriceFetchError
from decimal import Decimal

if TYPE_CHECKING:
    from py_portfolio_index.portfolio_providers.base_portfolio import BaseProvider


class Money(BaseModel):
    value: Union[Decimal, int, float, "Money"]
    currency: Currency = Currency.USD

    @validator("value", pre=True)
    def coerce_to_decimal(cls, v):
        if isinstance(v, int):
            return Decimal(v)
        if isinstance(v, Money):
            # TODO convert this
            return Decimal(v.value)
        return v

    def __str__(self):
        return f"{self.currency.value}{self.value}"

    def __repr__(self):
        return str(self)

    @classmethod
    def parse(cls, val) -> "Money":
        from py_portfolio_index.config import Config

        currency = Config.default_currency
        if isinstance(val, Money):
            return val
        elif isinstance(val, (Decimal, float, int)):
            return Money(value=Decimal(val), currency=currency)
        elif isinstance(val, str):
            for c in Currency:
                if c.name in val:
                    val = val.replace(c.name, "")
                    currency = c
            return Money(value=Decimal(val), currency=currency)
        raise ValueError(f"Invalid input to Money type {type(val)} {val}")

    def _cmp_helper(self, other):
        if isinstance(other, Money):
            if other.currency != self.currency:
                raise ValueError("Currency conversions not supported")
            return other.value
        return other

    def __eq__(self, other):
        return self.value == self._cmp_helper(other)

    def __ne__(self, other):
        return self.value != self._cmp_helper(other)

    def __gt__(self, other):
        return self.value > self._cmp_helper(other)

    def __ge__(self, other):
        return self.value >= self._cmp_helper(other)

    def __lt__(self, other):
        return self.value < self._cmp_helper(other)

    def __le__(self, other):
        return self.value <= self._cmp_helper(other)

    # sum starts with 0
    def __radd__(self, other) -> "Money":
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def __add__(self, other) -> "Money":
        return Money(value=self.value + self._cmp_helper(other), currency=self.currency)

    def __sub__(self, other) -> "Money":
        return Money(value=self.value - self._cmp_helper(other), currency=self.currency)

    def __mul__(self, other) -> "Money":
        return Money(value=self.value * self._cmp_helper(other), currency=self.currency)

    def __truediv__(self, other):
        return Money(value=self.value / self._cmp_helper(other), currency=self.currency)

    def __float__(self):
        return float(self.value)

    def __int__(self):
        return int(self.value)

    def __abs__(self):
        return Money(value=abs(self.value), currency=self.currency)

    def __round__(self, n=None):
        return Money(value=Decimal(round(self.value, n)), currency=self.currency)


class IdealPortfolioElement(BaseModel):
    ticker: str
    weight: Decimal


class IdealPortfolio(BaseModel):
    holdings: List[IdealPortfolioElement]
    source_date: Optional[date] = Field(default_factory=date.today)

    def _reweight_portfolio(self):
        weights: Decimal = sum([item.weight for item in self.holdings])

        scaling_factor = Decimal(1) / weights

        for item in self.holdings:
            item.weight = item.weight * scaling_factor

    def exclude(self, exclusion_list: List[str]):
        reweighted = []
        excluded = Decimal(0.0)
        for ticker in exclusion_list:
            reweighted.append(ticker)
            for item in self.holdings:
                if item.ticker == ticker:
                    excluded += item.weight
                    item.weight = Decimal(0.0)

        self.holdings = [
            item for item in self.holdings if item.ticker not in [reweighted]
        ]
        self._reweight_portfolio()
        Logger.info(
            f"Set the following stocks to weight 0 {reweighted}. Total value excluded {excluded}."
        )
        return self

    def reweight(
        self,
        ticker_list: List[str],
        weight: Union[Decimal, float],
        min_weight: Union[Decimal, float] = Decimal(0.005),
    ):
        cweight = Decimal(weight)
        cmin_weight = Decimal(min_weight)
        reweighted = []
        total_value = Decimal(0)
        for ticker in ticker_list:
            found = False
            for item in self.holdings:
                if item.ticker == ticker:
                    total_value += item.weight * cweight
                    item.weight = item.weight * cweight
                    reweighted.append(ticker)
                    found = True
            if not found:
                reweighted.append(ticker)
                total_value += cmin_weight
                self.holdings.append(
                    IdealPortfolioElement(ticker=ticker, weight=cmin_weight)
                )
        self._reweight_portfolio()
        Logger.info(
            f"modified the following by weight {cweight} {reweighted}. Total value modified {total_value}."
        )
        return self

    def reweight_to_present(self, provider: "BaseProvider"):
        imaginary_base = Decimal(1_000_000)
        values = {}
        for item in self.holdings:
            try:
                source_price = provider.get_instrument_price(
                    item.ticker, self.source_date
                )
            except PriceFetchError:
                source_price = provider.get_instrument_price(item.ticker)
            today_price = provider.get_instrument_price(item.ticker)
            if not source_price or not today_price:
                # if we couldn't get a historical price
                # keep the value the same
                values[item.ticker] = imaginary_base * item.weight
                continue
            source_shares = imaginary_base * item.weight / source_price

            today_value = today_price * source_shares
            values[item.ticker] = today_value
        today_value = Decimal(sum(values.values()))

        for item in self.holdings:
            new_weight = values[item.ticker] / today_value
            item.weight = new_weight
        self._reweight_portfolio()

    def produce_tear_sheet_from_date(self, datetime: datetime):
        raise NotImplementedError
        import pyfolio

        columns = ["date"] + [item.ticker for item in self.holdings]
        values = [datetime] + [item.value for item in self.holdings]
        df = DataFrame([values], columns=columns)
        df.set_index(keys=["date"], drop=True)
        pyfolio.create_simple_tear_sheet(positions=df, live_start_date=datetime)


class RealPortfolioElement(IdealPortfolioElement):
    ticker: str
    units: Decimal
    value: Money
    weight: Decimal = Decimal(0.0)

    @validator("value", pre=True)
    def value_coercion(cls, v) -> Money:
        return Money.parse(v)


class RealPortfolio(IdealPortfolio):
    holdings: List[RealPortfolioElement]  # type: ignore

    @property
    def _index(self):
        return {val.ticker: val for val in self.holdings}

    def get_holding(self, ticker: str):
        return self._index.get(ticker)

    @property
    def value(self) -> Money:
        values: List[Money] = [item.value for item in self.holdings]
        return Money(value=sum(values))

    def _reweight_portfolio(self):
        value = self.value
        for item in self.holdings:
            item.weight = Decimal(item.value / value.value)

    def add_holding(self, holding: RealPortfolioElement):
        existing = self._index.get(holding.ticker)
        if existing:
            existing.value += holding.value
            existing.units += holding.units
        if not existing:
            self.holdings.append(holding)
        self._reweight_portfolio()

    def __add__(self, other):
        if isinstance(other, RealPortfolioElement):
            self.add_holding(other)
        elif isinstance(other, RealPortfolio):
            for item in other.holdings:
                self.add_holding(item)
        else:
            raise ValueError
        return self



import re
from decimal import Decimal
from time import sleep
from datetime import date, datetime
import pandas as pd
from typing import Optional

from py_portfolio_index.constants import Logger
from py_portfolio_index.models import RealPortfolio, RealPortfolioElement
from .base_portfolio import BaseProvider
from py_portfolio_index.exceptions import PriceFetchError
from functools import lru_cache
from os import environ

FRACTIONAL_SLEEP = 60


def nearest_value(all_historicals, pivot) -> Optional[dict]:
    filtered = [z for z in all_historicals if z]
    if not filtered:
        return None
    return min(
        filtered,
        key=lambda x: abs(
            datetime.strptime(x["begins_at"], "%Y-%m-%dT%H:%M:%SZ").date() - pivot
        ),
    )


class RobinhoodProvider(BaseProvider):
    """Provider for interacting with Robinhood portfolios.

    Requires username and password.

    """

    def __init__(self, username: str | None = None, password: str | None = None):
        import robin_stocks.robinhood as r

        if not username:
            username = environ.get("ROBINHOOD_USERNAME", None)
        if not password:
            password = environ.get("ROBINHOOD_PASSWORD", None)
        if not (username and password):
            raise ValueError(
                "Must provide username and password arguments or set environment variables ROBINHOOD_USERNAME and ROBINHOOD_PASSWORD "
            )
        self._provider = r
        BaseProvider.__init__(self)
        self._provider.login(username=username, password=password)

    @lru_cache(maxsize=None)
    def _get_instrument_price(
        self, ticker: str, at_day: Optional[date] = None
    ) -> Optional[Decimal]:
        if at_day:
            historicals = self._provider.get_stock_historicals(
                [ticker], interval="day", span="year", bounds="regular"
            )
            closest = nearest_value(historicals, at_day)
            if closest:
                return Decimal(closest["high_price"])
            raise PriceFetchError(
                f"No historical data found for ticker {ticker} on date {at_day.isoformat()}"
            )
        else:
            quotes = self._provider.get_quotes([ticker])
            if not quotes[0]:
                return None
            return Decimal(quotes[0]["ask_price"])

    def buy_instrument(self, ticker: str, qty: Decimal):
        float_qty = float(qty)
        output = self._provider.order_buy_fractional_by_quantity(ticker, float_qty)
        msg = output.get("detail")
        if msg and "throttled" in msg:
            m = re.search("available in ([0-9]+) seconds", msg)
            if m:
                found = m.group(1)
                t = int(found)
            else:
                t = 30

            Logger.info(f"RH error: was throttled! Sleeping {t}")
            sleep(t)
            output = self.buy_instrument(ticker=ticker, qty=qty)
        elif msg and "Too many requests for fractional orders" in msg:
            Logger.info(
                f"RH error: was throttled on fractional orders! Sleeping {FRACTIONAL_SLEEP}"
            )
            sleep(FRACTIONAL_SLEEP)
            output = self.buy_instrument(ticker=ticker, qty=qty)
        if not output.get("id"):
            Logger.error(msg)
            raise ValueError(msg)
        return True

    def get_unsettled_instruments(self):
        orders = self._provider.get_all_open_stock_orders()
        for item in orders:
            item["symbol"] = self._provider.get_symbol_by_url(item["instrument"])
        return set(item["symbol"] for item in orders)

    def get_holdings(self):
        my_stocks = self._provider.build_holdings()

        df = pd.DataFrame(my_stocks)
        df = df.T
        df["ticker"] = df.index
        df = df.reset_index(drop=True)
        df.sort_values(by=["percentage"], inplace=True)
        out = [
            RealPortfolioElement(
                ticker=row.ticker,
                units=Decimal(row.quantity),
                value=Decimal(row.equity),
                weight=Decimal(row.percentage) / 100,
            )
            for row in df.itertuples()
        ]
        return RealPortfolio(holdings=out)



from enum import IntEnum, Enum

class Provider(str, Enum):
    ALPACA = "alpaca"
    ROBINHOOD = "robinhood"

class PurchaseStrategy(IntEnum):
    CHEAPEST_FIRST = 1
    LARGEST_DIFF_FIRST = 2


class RoundingStrategy(int, Enum):
    CLOSEST = 1
    FLOOR = 2
    CEILING = 3


class Currency(str, Enum):
    USD = "$"
    EURO = "€"
    POUND = "£"



from dataclasses import dataclass
from typing import Optional, Dict, Union
from decimal import Decimal

from py_portfolio_index.common import print_per
from py_portfolio_index.constants import Logger
from py_portfolio_index.enums import PurchaseStrategy
from py_portfolio_index.models import Money
from .models import IdealPortfolio, RealPortfolio


@dataclass
class ComparisonResult:
    ticker: str
    model: Decimal
    comparison: Decimal
    actual: Money

    @property
    def diff(self):
        return self.model - Decimal(self.comparison)


def compare_portfolios(
    real: RealPortfolio,
    ideal: IdealPortfolio,
    buy_order=PurchaseStrategy.LARGEST_DIFF_FIRST,
    target_size: Optional[Union[Decimal, int]] = None,
):
    output: Dict[str, ComparisonResult] = {}
    diff = Decimal(0.0)
    selling = Decimal(0.0)
    buying = Decimal(0.0)
    target_value: Money = Money(value=Decimal(target_size)) if target_size else real.value
    for value in ideal.holdings:
        comparison = real.get_holding(value.ticker)
        if not comparison:
            percentage = Decimal(0.0)
            actual_value = Money.parse("0.0")
        else:
            percentage = Decimal((comparison.value / target_value).value)
            actual_value = comparison.value
        output[value.ticker] = ComparisonResult(
            ticker=value.ticker,
            model=value.weight,
            comparison=percentage,
            actual=actual_value,
        )
        _diff = Decimal(value.weight) - percentage
        diff += abs(_diff)
        if _diff < 0:
            selling += abs(_diff)
        else:
            buying += abs(_diff)

    Logger.info(
        f"Total portfolio % delta {print_per(diff)}. Overweight {print_per(selling)}, underweight {print_per(buying)}"
    )
    if buy_order == PurchaseStrategy.LARGEST_DIFF_FIRST:
        diff_output: Dict[str, ComparisonResult] = {
            k: v for k, v in sorted(output.items(), key=lambda item: -abs(item[1].diff))
        }
    elif buy_order == PurchaseStrategy.CHEAPEST_FIRST:
        diff_output = {
            k: v for k, v in sorted(output.items(), key=lambda item: abs(item[1].diff))
        }
    else:
        raise ValueError("Invalid purchase strategy")
    to_purchase = {}
    to_sell = {}
    for key, diffvalue in diff_output.items():
        if diffvalue.diff == 0:
            continue
        elif diffvalue.diff < 0:
            diff_text = "Overweight"
            to_sell[key] = (
                target_value * diffvalue.comparison - target_value * diffvalue.model
            )
        else:
            diff_text = "Underweight"
            to_purchase[key] = (
                target_value * diffvalue.model - target_value * diffvalue.comparison
            )
            # to_purchase.append(key)
        Logger.info(
            f"{diff_text} {key}, {print_per(diffvalue.model)} target vs {print_per(diffvalue.comparison)} actual. Should be {target_value * diffvalue.model}, is {diffvalue.actual}"
        )
    return to_purchase, to_sell



from typing import Set
import re
from datetime import date
from decimal import Decimal 
from pydantic import BaseModel, Field
from pathlib import Path

from py_portfolio_index.models import IdealPortfolioElement, IdealPortfolio

QUARTER_TO_MONTH = {1: 1, 2: 4, 3: 7, 4: 10}


def parse_date_from_name(input: str)-> date | None:
    components = input.lower().split("_")
    year = None
    quarter = None
    for item in components:
        if re.match("[0-9]{4}", item):
            year = int(item)
        if re.match("q[1-4]", item):
            quarter = int(item[1])
    if not year or not quarter:
        return None

    return date(year=year, month=QUARTER_TO_MONTH[quarter], day=1)


class IndexInventory(BaseModel):
    keys: Set[str] = Field(exclude=True)
    base: Path = Field(exclude=True)
    loaded: dict[str, IdealPortfolio] = Field(default_factory = dict)

    @classmethod
    def from_path(cls, path):
        path = Path(path)
        if path.is_file():
            path = path.parent
        keys = []
        for f in path.iterdir():
            try:
                if f.suffix == ".csv":
                    keys.append(f.stem)
            except FileNotFoundError:
                pass
        return IndexInventory(keys = keys, base = path)

    def __getitem__(self, item: str) -> IdealPortfolio:
        if item in self.keys:
            values = self.loaded.get(item, None)
            if values:
                return values
            values = self.get_values(item)
            self.loaded[item] = values
            return values
        raise KeyError(item)

    def get_values(self, item: str) -> IdealPortfolio:
        out = []
        with open(self.base / f"{item}.csv") as f:
            contents = f.read()
            for row in contents.split("\n"):
                ticker, weight = row.split(",", 1)
                out.append(IdealPortfolioElement(ticker=ticker, weight=Decimal(weight)))
        start_date = parse_date_from_name(item)
        if start_date:
            return IdealPortfolio(holdings=out, source_date=start_date)
        return IdealPortfolio(holdings=out)



import random
from typing import List, Dict, Optional
from datetime import date
from py_portfolio_index.models import (
    RealPortfolio,
    RealPortfolioElement,
)
from py_portfolio_index.models import Money
from decimal import Decimal
from .base_portfolio import BaseProvider


class FixedGen:
    def __init__(self, value: float):
        self.value = value

    def get(self):
        return self.value


class RandGen:
    def __init__(self, seed: Optional[int] = None):
        random.seed(seed)

    def get(self):
        return random.randint(150, 10000) / 100


class LocalDictProvider(BaseProvider):
    def __init__(
        self,
        holdings: List[RealPortfolioElement],
        price_dict: Optional[Dict[str, Decimal]] = None,
        default_price_gen=RandGen,
    ):
        BaseProvider.__init__(self)
        self._price_dict = price_dict or {}
        self._portfolio = RealPortfolio(holdings=holdings)
        self.default_price_gen = default_price_gen()

    def _get_instrument_price(self, ticker: str, at_day: Optional[date] = None):
        value = self._price_dict.get(ticker)
        if not value:
            nvalue = self.default_price_gen.get()
            self._price_dict[ticker] = nvalue
            return nvalue
        return value

    def buy_instrument(self, ticker: str, qty: Decimal):
        price = self.get_instrument_price(ticker)
        if not price:
            raise ValueError("No available price for this instrument")
        self._portfolio += RealPortfolioElement(
            ticker=ticker, units=qty, value=Money(value=qty * price)
        )

    def get_holdings(self):
        return self._portfolio
